package AseWriter;

# read ogre mesh xml file and write 3ds MAX ase format
#
# Written by Martin Correll for GameCat, contact: mac AT gamecat DOT de
# Licence: Perl Artistic Licence (http://www.perl.com/pub/a/language/misc/Artistic.html)

use strict;
use XML::Simple;
use Data::Dumper;

use constant DEBUG => 1;


sub new {
	my ($class, %args) = @_;
	my $self = {};
	foreach ( keys %args ) {
		$self->{$_} = $args{$_};
	}
	bless($self, $class);
	return($self);
}


sub read_xml {
	my ($self, $mesh) = @_;

	warn "converting $mesh to ase...\n" if DEBUG;
	my $ref = XMLin( $mesh, ForceArray => ['submesh', 'vertexbuffer'] );

	my $vertex_count		= 0;
	my $vertex_index_offset	= 0;
	my $face_count			= 0;
	my @submeshes;
	my @normals;

	foreach my $submesh ( @{$ref->{submeshes}->{submesh}} ) {
		my %sub;
		$sub{material} = $submesh->{material};

		$vertex_index_offset = $vertex_count;

		foreach my $vertexbuffer ( @{ $submesh->{geometry}->{vertexbuffer} } ) {
			if ( $vertexbuffer->{positions} ) {
				foreach my $vertex ( @{ $vertexbuffer->{vertex} } ) {
					push( @{$sub{positions}}, [$vertex->{position}->{x}, - $vertex->{position}->{z}, $vertex->{position}->{y}] );
					$vertex_count++;
				}
			}
			if ( $vertexbuffer->{normals} ) {
				foreach my $vertex ( @{ $vertexbuffer->{vertex} } ) {
					push( @{$sub{normal}}, [$vertex->{normal}->{x}, - $vertex->{normal}->{z}, $vertex->{normal}->{y}] );
					push( @normals, [$vertex->{normal}->{x}, - $vertex->{normal}->{z}, $vertex->{normal}->{y}] );
				}
			}
			if ( $vertexbuffer->{texture_coords} ) {
				foreach my $vertex ( @{ $vertexbuffer->{vertex} } ) {
					push( @{$sub{texcoord}}, [$vertex->{texcoord}->{u}, 1 - $vertex->{texcoord}->{v}] );
				}
			}
		}

		foreach my $face ( @{ $submesh->{faces}->{face} } ) {
			push( @{$sub{faces}}, [$face->{v1} + $vertex_index_offset, $face->{v2} + $vertex_index_offset, $face->{v3} + $vertex_index_offset] );
			$face_count++;
		}

		push( @submeshes, \%sub );
	}

	warn "vertex_count = $vertex_count\n" if DEBUG;
	warn "face_count = $face_count\n" if DEBUG;

	$self->{vertex_count}	= $vertex_count;
	$self->{face_count}		= $face_count;

	$self->{submeshes}		= \@submeshes;
	$self->{normals}		= \@normals;
}


sub write {
	my $self = shift;

	$self->write_header();
	$self->write_material_header();
	$self->write_materials( $self->{ogre_materials} );
	$self->write_mesh_header();
	$self->write_vertices();
	$self->write_faces();
	$self->write_texture_coords();
	$self->write_textured_faces();
	$self->write_normals();
	$self->write_footer();
}


sub write_header {
	my $self = shift;
	print <<EOStr;
*3DSMAX_ASCIIEXPORT	200
*COMMENT "created by mesh2ase.pl"
*SCENE {
	*SCENE_FILENAME ""
	*SCENE_FIRSTFRAME 0
	*SCENE_LASTFRAME 100
	*SCENE_FRAMESPEED 30
	*SCENE_TICKSPERFRAME 160
	*SCENE_BACKGROUND_STATIC 0.0000	0.0000	0.0000
	*SCENE_AMBIENT_STATIC 0.0000	0.0000	0.0000
}

EOStr

}


sub write_material_header {
	my $self = shift;

	print <<EOStr;
*MATERIAL_LIST {
	*MATERIAL_COUNT 1
	*MATERIAL 0 {
		*MATERIAL_NAME "Master"
		*MATERIAL_CLASS "Multi/Sub-Object"
		*MATERIAL_AMBIENT 0.5880	0.5880	0.5880
		*MATERIAL_DIFFUSE 0.5880	0.5880	0.5880
		*MATERIAL_SPECULAR 0.9000	0.9000	0.9000
		*MATERIAL_SHINE 0.1000
		*MATERIAL_SHINESTRENGTH 0.0000
		*MATERIAL_TRANSPARENCY 0.0000
		*MATERIAL_WIRESIZE 1.0000
EOStr

	print "\t\t*NUMSUBMTLS " . scalar @{$self->{submeshes}} . "\n";
}


sub write_materials {
	my ($self, $materials) = @_;

	my $material_count = 0;
	foreach my $sub (@{$self->{submeshes}}) {
		my $current_material = $materials->{ $sub->{material} };

		print <<EOStr;
		*SUBMATERIAL $material_count {
			*MATERIAL_NAME "$sub->{material}"
			*MATERIAL_CLASS "Standard"
EOStr

		unless ( exists $current_material ->{ambient} ) {
			foreach my $i (0..2) {
				$current_material->{ambient}->[ $i ] = "1.0";
			}
		}
		unless ( exists $current_material ->{diffuse} ) {
			foreach my $i (0..2) {
				$current_material->{diffuse}->[ $i ] = "1.0";
			}
		}

		print "\t\t\t*MATERIAL_AMBIENT " . join( "\t", @{ $current_material->{ambient} } ) . "\n";
		print "\t\t\t*MATERIAL_DIFFUSE " . join( "\t", @{ $current_material->{diffuse} } ) . "\n";

		print <<EOStr;
			*MATERIAL_SPECULAR 0.9000	0.9000	0.9000
			*MATERIAL_SHINE 0.1000
			*MATERIAL_SHINESTRENGTH 0.0000
			*MATERIAL_TRANSPARENCY 0.0000
			*MATERIAL_WIRESIZE 1.0000
			*MATERIAL_SHADING Blinn
			*MATERIAL_XP_FALLOFF 0.0000
			*MATERIAL_SELFILLUM 0.0000
			*MATERIAL_FALLOFF In
			*MATERIAL_XP_TYPE Filter
EOStr


		if ( exists $materials->{ $sub->{material} } ) {
			my $bitmap = $materials->{ $sub->{material} }->{texture_with_path};
			print <<EOStr;
			*MAP_DIFFUSE {
				*MAP_NAME "Map #1"
				*MAP_CLASS "Bitmap"
				*MAP_SUBNO 1
				*MAP_AMOUNT 1.0000
				*BITMAP "$bitmap"
				*MAP_TYPE Screen
				*UVW_U_OFFSET 0.0000
				*UVW_V_OFFSET 0.0000
				*UVW_U_TILING 1.0000
				*UVW_V_TILING 1.0000
				*UVW_ANGLE 0.0000
				*UVW_BLUR 1.0000
				*UVW_BLUR_OFFSET 0.0000
				*UVW_NOUSE_AMT 1.0000
				*UVW_NOISE_SIZE 1.0000
				*UVW_NOISE_LEVEL 1
				*UVW_NOISE_PHASE 0.0000
				*BITMAP_FILTER Pyramidal
			}
EOStr

		}

		print "\t\t}\n";
		$material_count++;
	}

	print "\t}\n";
	print "}\n";
}


sub write_mesh_header {
	my $self = shift;

	print <<EOStr;
*GEOMOBJECT {
	*NODE_NAME "ogrescene"
	*NODE_TM {
		*NODE_NAME "ogrescene"
		*INHERIT_POS 0 0 0
		*INHERIT_ROT 0 0 0
		*INHERIT_SCL 0 0 0
		*TM_ROW0 1.0000	0.0000	0.0000
		*TM_ROW1 0.0000	1.0000	0.0000
		*TM_ROW2 0.0000	0.0000	1.0000
		*TM_ROW3 6.6421	-7.3801	0.0000
		*TM_POS 6.6421	-7.3801	0.0000
		*TM_ROTAXIS 0.0000	0.0000	0.0000
		*TM_ROTANGLE 0.0000
		*TM_SCALE 1.0000	1.0000	1.0000
		*TM_SCALEAXIS 0.0000	0.0000	0.0000
		*TM_SCALEAXISANG 0.0000
	}
EOStr

	print "\t*MESH {\n";
	print "\t\t*TIMEVALUE 0\n";
	print "\t\t*MESH_NUMVERTEX " . $self->{vertex_count} . "\n";
	print "\t\t*MESH_NUMFACES " . $self->{face_count} . "\n";
}


sub write_vertices {
	my $self = shift;

	print "\t\t*MESH_VERTEX_LIST {\n";
	my $vertex_id = 0;
	foreach my $sub ( @{$self->{submeshes}} ) {
		foreach my $pos ( @{$sub->{positions}} ) {
			print "\t\t\t*MESH_VERTEX    " . $vertex_id . "\t" . $pos->[0] . "\t" .  $pos->[1] . "\t" .  $pos->[2] . "\n";
			$vertex_id++;
		}
	}
	print "\t\t}\n";
}


sub write_faces {
	my $self = shift;

	print "\t\t*MESH_FACE_LIST {\n";
	my $face_id = 0;
	my $material_id = 0;
	foreach my $sub ( @{$self->{submeshes}} ) {
		foreach my $face ( @{$sub->{faces}} ) {
			print "\t\t\t*MESH_FACE\t$face_id:\tA:\t" . $face->[0] . " B:\t" . $face->[1] . " C:\t" . $face->[2]
				. " AB:\t1 BC:\t1 CA:\t0\t*MESH_SMOOTHING $material_id\t*MESH_MTLID $material_id\n";
			$face_id++;
		}
		$material_id++;
	}
	print "\t\t}\n";
}


sub write_texture_coords {
	my $self = shift;
	print "\t\t*MESH_NUMTVERTEX $self->{vertex_count}\n";
	print "\t\t*MESH_TVERTLIST {\n";
	my $tex_id = 0;
	foreach my $sub ( @{$self->{submeshes}} ) {
		foreach my $tex ( @{$sub->{texcoord}} ) {
			print "\t\t\t*MESH_TVERT $tex_id " . $tex->[0] . " " . $tex->[1] . " 0.0000\n";
			$tex_id++;
		}
	}
	print "\t\t}\n";
}


sub write_textured_faces {
	my $self = shift;

	print "\t\t*MESH_NUMTVFACES $self->{face_count}\n";
	print "\t\t*MESH_TFACELIST {\n";
	my $tface_id = 0;
	foreach my $sub ( @{$self->{submeshes}} ) {
		foreach my $face ( @{$sub->{faces}} ) {
			print "\t\t\t*MESH_TFACE $tface_id $face->[0] $face->[1] $face->[2]\n";
			$tface_id++;
		}
	}
	print "\t\t}\n";
}


sub write_normals {
	my $self = shift;

	print "\t\t*MESH_NORMALS {\n";
	my $face_id = 0;
	foreach my $sub ( @{$self->{submeshes}} ) {
		foreach my $face ( @{$sub->{faces}} ) {
			my $x0 = $self->{normals}[ $face->[0] ]->[0];
			my $x1 = $self->{normals}[ $face->[1] ]->[0];
			my $x2 = $self->{normals}[ $face->[2] ]->[0];

			my $y0 = $self->{normals}[ $face->[0] ]->[1];
			my $y1 = $self->{normals}[ $face->[1] ]->[1];
			my $y2 = $self->{normals}[ $face->[2] ]->[1];

			my $z0 = $self->{normals}[ $face->[0] ]->[2];
			my $z1 = $self->{normals}[ $face->[1] ]->[2];
			my $z2 = $self->{normals}[ $face->[2] ]->[2];

			# "calculate" face normals
			my $faceX = ( $x0 + $x1 + $x2 ) / 3;
			my $faceY = ( $y0 + $y1 + $y2 ) / 3;
			my $faceZ = ( $z0 + $z1 + $z2 ) / 3;

			print "\t\t\t*MESH_FACENORMAL $face_id $faceX $faceY $faceZ\n";
			print "\t\t\t*MESH_VERTEXNORMAL " . $face->[0] . " " . $x0 . " " . $y0 . " " . $z0 . "\n";
			print "\t\t\t*MESH_VERTEXNORMAL " . $face->[1] . " " . $x1 . " " . $y1 . " " . $z1 . "\n";
			print "\t\t\t*MESH_VERTEXNORMAL " . $face->[2] . " " . $x2 . " " . $y2 . " " . $z2 . "\n";

			$face_id++;
		}
	}
	print "\t\t}\n";
}


sub write_footer {
	my $self = shift;

	print "\t}\n";
	print <<EOStr;
	*PROP_MOTIONBLUR 0
	*PROP_CASTSHADOW 1
	*PROP_RECVSHADOW 1
	*MATERIAL_REF 0
}

EOStr

}

1;
